#include <string.h> /* for strcmp() */
#include <gdk/gdkx.h> /* for gdk_x11_window_set_user_time() */

#include "tilda.h"
#include "tilda-controller.h"
#include "tilda-window.h"
#include "tilda-window-dbus-glue.h"
#include "tomboykeybinder.h"

/**
 * Find the TildaTerminal corresponding to the currently selected
 * tab in self->notebook. This could go away if TildaTerminal were
 * a proper subclass of GtkWidget.
 */
static TildaTerminal *
tilda_window_find_current_terminal (TildaWindow *self)
{
	debug_enter();
	debug_assert (TILDA_IS_WINDOW(self));

	gint i;
	TildaTerminal *ret;
	gint current_page = gtk_notebook_get_current_page (GTK_NOTEBOOK(self->notebook));
	GtkWidget *box = gtk_notebook_get_nth_page (GTK_NOTEBOOK(self->notebook), current_page);

	for (i=0; i<self->terms->len; ++i)
	{
		ret = g_ptr_array_index (self->terms, i);

		if (ret->hbox == box)
			return ret;
	}

	debug_printf ("ERROR: unable to find current terminal!\n");
	return NULL;
}

static gint
tilda_window_find_next_free_terminal_number (TildaWindow *self)
{
	debug_enter ();
	debug_assert (TILDA_IS_WINDOW(self));

	gint i, j;
	gboolean found;

	for (i=0; i<INT_MAX; ++i)
	{
		found = FALSE;

		for (j=0; j<self->terms->len; ++j)
		{
			TildaTerminal *tt = g_ptr_array_index (self->terms, j);

			if (tt->number == i)
			{
				found = TRUE;
				break;
			}
		}

		if (!found)
			return i;
	}

	return 0;
}

/**
 * Clean up and remove self completely from the program
 *
 * Should only be used by DBus...
 */
gboolean
tilda_window_close (TildaWindow *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	tilda_controller_remove_window (TILDA_CONTROLLER(self->controller), self->number);

	return TRUE;
}

gboolean
tilda_window_add_terminal (TildaWindow *self)
{
	debug_enter ();
	debug_assert (TILDA_IS_WINDOW(self));

	gint number;
	TildaTerminal *tt;

	number = tilda_window_find_next_free_terminal_number (self);
	tt = g_object_new (TILDA_TYPE_TERMINAL,
					   "number", number,
					   "parent-window", self,
					   NULL);
	g_ptr_array_add (self->terms, tt);

	GtkWidget *label = gtk_label_new ("Tilda");
	gint index = gtk_notebook_prepend_page (GTK_NOTEBOOK(self->notebook), tt->hbox, label);
	gtk_notebook_set_tab_label_packing (GTK_NOTEBOOK(self->notebook), tt->hbox, TRUE, TRUE, GTK_PACK_END);
	gtk_notebook_set_current_page (GTK_NOTEBOOK(self->notebook), index);

	if (gtk_notebook_get_n_pages (GTK_NOTEBOOK(self->notebook)) > 1)
		gtk_notebook_set_show_tabs (GTK_NOTEBOOK(self->notebook), TRUE);

	/* Focus the VTE Terminal */
	gtk_widget_grab_focus (tt->vte_term);

	return TRUE;
}

/**
 * Remove the TildaTerminal with the given number from the given
 * TildaWindow.
 *
 * Return: TRUE on success, FALSE otherwise.
 */
gboolean
tilda_window_remove_terminal (TildaWindow *self, gint terminal_number)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));
	debug_assert (terminal_number >= 0);

	gint i;

	for (i=0; i<self->terms->len; ++i)
	{
		TildaTerminal *tt = g_ptr_array_index (self->terms, i);

		if (tt->number == terminal_number)
		{
			gint notebook_index = gtk_notebook_page_num (GTK_NOTEBOOK(self->notebook), tt->hbox);

			/* Make sure the index was valid */
			if (notebook_index == -1)
			{
				debug_printf ("ERROR: Bad Notebook Tab\n");
				return FALSE;
			}

			/* Actually remove the terminal */
			gtk_notebook_remove_page (GTK_NOTEBOOK (self->notebook), notebook_index);

			/* We should hide the tabs if there is only one tab left */
			if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (self->notebook)) == 1)
				gtk_notebook_set_show_tabs (GTK_NOTEBOOK (self->notebook), FALSE);

			/* Remove the term from our lists, then free it */
			g_ptr_array_remove_fast (self->terms, tt);
			g_object_unref (G_OBJECT(tt));

			/* With no pages left, it's time to remove this window */
			if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (self->notebook)) < 1)
			{
				debug_printf ("no terminals left, closing window %d\n", self->number);
				tilda_controller_remove_window (TILDA_CONTROLLER(self->controller), self->number);
			}

			/* Leave the loop, we're done */
			break;
		}
	}

	return TRUE;
}

/**
 * This sets up the given TildaWindow for the capability of real
 * transparency, if the X server is capable of it. */
static void
tilda_window_setup_real_transparency (TildaWindow *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	GdkScreen *screen;
	GdkColormap *colormap;

	screen = gtk_widget_get_screen (GTK_WIDGET(self->window));
	colormap = gdk_screen_get_rgba_colormap (screen);

	/* If possible, set the RGBA colormap so VTE can use real alpha
	 * channels for transparency. */
	if (colormap != NULL && gdk_screen_is_composited (screen))
	{
		gtk_widget_set_colormap (GTK_WIDGET(self->window), colormap);
		self->have_real_transparency = TRUE;
		return;
	}

	self->have_real_transparency = FALSE;
}

/* Center the given TildaWindow in the horizontal axis */
static void
tilda_window_center_horizontally (TildaWindow *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	const gint screen_center = gdk_screen_width() / 2;
	const gint tilda_center  = self->width / 2;
	const gint center_coord  = screen_center - tilda_center;

	g_object_set (G_OBJECT(self), "x-position", center_coord, NULL);
}

/* Center the given TildaWindow in the vertical axis */
static void
tilda_window_center_vertically (TildaWindow *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	const gint screen_center = gdk_screen_height() / 2;
	const gint tilda_center  = self->height / 2;
	const gint center_coord  = screen_center - tilda_center;

	g_object_set (G_OBJECT(self), "y-position", center_coord, NULL);
}

static void
tilda_window_keybinding_cb (const gchar *keystr, gpointer data)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(data));

	TildaWindow *self = TILDA_WINDOW(data);
	TildaTerminal *tt;

	/* This call sets the X11 window property _NET_WM_USER_TIME, which GTK+ normally
	 * sets for us. However, because this callback is activated via a global keybinding,
	 * we see the event before GDK / GTK+ does. Therefore, to get the focus, we must
	 * set the property ourselves. */
	gdk_x11_window_set_user_time (GTK_WIDGET(self->window)->window,
								  tomboy_keybinder_get_current_event_time());

	switch (self->state)
	{
		case WINDOW_UP: /* Pull the window up */

			/* Bugfix: having this here keeps the tilda window from being
			 * hidden if you turn off "stick", pull it down on workspace 1,
			 * switch to workspace 2, then pull it up and back down. Without
			 * this, something in metacity (at least) hides the window. Stupid. */
			gtk_window_deiconify (GTK_WINDOW(self->window));

			/* Re-set the window properties that do not linger after hiding the
			 * window. I know this looks stupid, but it keeps all of the state-
			 * changing code in the place it belongs: the property-setting code. */
			g_object_set (G_OBJECT(self),
					"keep-above", self->keep_above,
					"stick", self->stick,
					NULL);
			gtk_widget_show (GTK_WIDGET(self->window));

			/* Focusing the term here works perfectly, near as I can tell */
			tt = tilda_window_find_current_terminal (self);
			gtk_widget_grab_focus (GTK_WIDGET(tt->vte_term));

			self->state = WINDOW_DOWN;
			break;

		case WINDOW_DOWN: /* Pull the window up */

			gtk_widget_hide (GTK_WIDGET(self->window));

			self->state = WINDOW_UP;
			break;

		default:
			debug_printf ("ERROR: Window is in a bad state!\n");

			/* Pretend we're down, for good measure.... */
			self->state = WINDOW_DOWN;
			break;
	}
}

/**
 * Attempt to bind the new_key to show this window.
 *
 * Return: TRUE if successful, FALSE otherwise.
 */
static gboolean
tilda_window_try_to_bind_key (TildaWindow *self, const gchar *new_key)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	gboolean ret = FALSE;

	/* Make sure the new key is not null in any way */
	if (new_key == NULL || strcmp("", new_key) == 0)
		return FALSE;

	/* Check that no other windows are using the key */
	// FIXME: there should be a hidden option to disable this. Maybe some people want
	// to have logs in two Tildas, and just show them with one key. Crazy...
	if (tilda_controller_global_key_in_use(TILDA_CONTROLLER(self->controller), new_key))
		return FALSE;

	/* Unbind if we were set */
	if (self->key)
		tomboy_keybinder_unbind (self->key, tilda_window_keybinding_cb);

	ret = tomboy_keybinder_bind (new_key, tilda_window_keybinding_cb, self);

	/* If it was successful, update the self->key variable and be done with it */
	if (ret)
	{
		g_free (self->key);
		self->key = g_strdup (new_key);
		return TRUE;
	}

	g_printerr (_("Bind key '%s' failed. Reverting to original keybinding\n"), self->key);

	/* Not successful, so rebind the old key, and return FALSE */
	if (self->key != NULL && strcmp("",self->key) != 0)
	{
		ret = tomboy_keybinder_bind (self->key, tilda_window_keybinding_cb, self);

		/* Check that it went ok */
		if (!ret)
			g_printerr (_("Unable to re-bind original key '%s'. Oh shit...\n"), self->key);
	}
	else
		g_printerr (_("No original key to revert to!\n"));

	return FALSE;
}

static void
tilda_window_dbus_register_object (TildaWindow *self)
{
	debug_enter  ();
	debug_assert (TILDA_IS_WINDOW(self));

	gchar *object_path;

	// Register this object with DBus
	object_path = g_strdup_printf ("/net/sourceforge/Tilda/Window%d", self->number);
	dbus_g_connection_register_g_object (dbus_connection, object_path, G_OBJECT(self));
	g_free (object_path);
}

/*******************************************************************************
 * ALL GOBJECT STUFF BELOW PLEASE
 ******************************************************************************/

static GObjectClass *parent_class = NULL;

enum tilda_window_properties {
	TILDA_WINDOW_NUMBER = 1,
	TILDA_WINDOW_CONTROLLER,

	TILDA_WINDOW_KEY,

	TILDA_WINDOW_HEIGHT,
	TILDA_WINDOW_WIDTH,
	TILDA_WINDOW_X_POSITION,
	TILDA_WINDOW_Y_POSITION,

	TILDA_WINDOW_TAB_POSITION,
	TILDA_WINDOW_ANIMATION_ORIENTATION,
	TILDA_WINDOW_ANIMATION_DELAY,

	TILDA_WINDOW_KEEP_ABOVE,
	TILDA_WINDOW_SKIP_TASKBAR_HINT,
	TILDA_WINDOW_STICK,
	TILDA_WINDOW_HIDDEN_AT_START,
	TILDA_WINDOW_CENTERED_HORIZONTALLY,
	TILDA_WINDOW_CENTERED_VERTICALLY,

	TILDA_WINDOW_HAVE_REAL_TRANSPARENCY,
};

static void
tilda_window_instance_init (GTypeInstance *instance,
							gpointer       g_class)
{
	debug_enter ();

	TildaWindow *self = (TildaWindow *) instance;
	self->dispose_has_run = FALSE;

	/* Initialize all properties */
	self->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	self->notebook = gtk_notebook_new ();
	self->terms = g_ptr_array_new ();

	/* Somewhat of a "poison" value, incase we don't set this */
	self->number = 0xdeadbeef;
	self->controller = NULL;

	self->state = WINDOW_UP;
}

static void
tilda_window_set_property (GObject      *object,
						   guint         property_id,
						   const GValue *value,
						   GParamSpec   *pspec)
{
	TildaWindow *self = (TildaWindow *) object;

	switch (property_id) {

		case TILDA_WINDOW_NUMBER:
			self->number = g_value_get_int (value);
			debug_printf ("window number: %d\n", self->number);
			break;

		case TILDA_WINDOW_CONTROLLER:
			self->controller = g_value_get_pointer (value);
			debug_printf ("window controller: 0x%x\n", self->controller);
			break;

		case TILDA_WINDOW_KEY:
			tilda_window_try_to_bind_key (self, g_value_get_string (value));
			debug_printf ("window key %s\n", self->key);
			break;

		case TILDA_WINDOW_HEIGHT:
			self->height = g_value_get_int (value);
			gtk_widget_set_size_request (self->window, self->width, self->height);
			gtk_window_resize (GTK_WINDOW(self->window), self->width, self->height);
			debug_printf ("window height: %d\n", self->height);
			break;

		case TILDA_WINDOW_WIDTH:
			self->width = g_value_get_int (value);
			gtk_widget_set_size_request (self->window, self->width, self->height);
			gtk_window_resize (GTK_WINDOW(self->window), self->width, self->height);
			debug_printf ("window width: %d\n", self->width);
			break;

		case TILDA_WINDOW_X_POSITION:
			self->x_position = g_value_get_int (value);
			gtk_window_move (GTK_WINDOW(self->window), self->x_position, self->y_position);
			debug_printf ("window x position: %d\n", self->x_position);
			break;

		case TILDA_WINDOW_Y_POSITION:
			self->y_position = g_value_get_int (value);
			gtk_window_move (GTK_WINDOW(self->window), self->x_position, self->y_position);
			debug_printf ("window y position: %d\n", self->y_position);
			break;

		case TILDA_WINDOW_TAB_POSITION:
			self->tab_position = g_value_get_int (value);
			gtk_notebook_set_tab_pos (GTK_NOTEBOOK(self->notebook), self->tab_position);
			debug_printf ("window tab position: %d\n", self->tab_position);
			break;

		case TILDA_WINDOW_ANIMATION_ORIENTATION:
			self->animation_orientation = g_value_get_int (value);
			debug_printf ("window animation orientation: %d\n", self->animation_orientation);
			break;

		case TILDA_WINDOW_ANIMATION_DELAY:
			self->animation_delay = g_value_get_int (value);
			debug_printf ("window animation delay: %d\n", self->animation_delay);
			break;

		case TILDA_WINDOW_KEEP_ABOVE:
			self->keep_above = g_value_get_boolean (value);
			gtk_window_set_keep_above (GTK_WINDOW(self->window), self->keep_above);
			debug_printf ("window keep above: %d\n", self->keep_above);
			break;

		case TILDA_WINDOW_SKIP_TASKBAR_HINT:
			self->skip_taskbar_hint = g_value_get_boolean (value);
			gtk_window_set_skip_taskbar_hint (GTK_WINDOW(self->window), self->skip_taskbar_hint);
			debug_printf ("window skip taskbar hint: %d\n", self->skip_taskbar_hint);
			break;

		case TILDA_WINDOW_STICK:
			self->stick = g_value_get_boolean (value);

			/* This is moderately ugly, but GTK+ does it this way... */
			self->stick ? gtk_window_stick (GTK_WINDOW(self->window))
						: gtk_window_unstick (GTK_WINDOW(self->window));
			debug_printf ("window stick: %d\n", self->stick);
			break;

		case TILDA_WINDOW_HIDDEN_AT_START:
			self->hidden_at_start = g_value_get_boolean (value);
			debug_printf ("window hidden at start: %d\n", self->hidden_at_start);
			break;

		case TILDA_WINDOW_CENTERED_HORIZONTALLY:
			self->centered_horizontally = g_value_get_boolean (value);
			if (self->centered_horizontally)
				tilda_window_center_horizontally (self);
			debug_printf ("window centered horizontally: %d\n", self->centered_horizontally);
			break;

		case TILDA_WINDOW_CENTERED_VERTICALLY:
			self->centered_vertically = g_value_get_boolean (value);
			if (self->centered_vertically)
				tilda_window_center_vertically (self);
			debug_printf ("window centered vertically: %d\n", self->centered_vertically);
			break;

		case TILDA_WINDOW_HAVE_REAL_TRANSPARENCY:
			self->have_real_transparency = g_value_get_boolean (value);
			debug_printf ("window have real transp: %d\n", self->have_real_transparency);
			break;

		default:
			/* We don't have this property */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static void
tilda_window_get_property (GObject    *object,
						   guint       property_id,
						   GValue     *value,
						   GParamSpec *pspec)
{
	TildaWindow *self = (TildaWindow *) object;

	switch (property_id) {

		case TILDA_WINDOW_NUMBER:
			g_value_set_int (value, self->number);
			break;

		case TILDA_WINDOW_CONTROLLER:
			g_value_set_pointer (value, self->controller);
			break;

		case TILDA_WINDOW_KEY:
			g_value_set_string (value, self->key);
			break;

		case TILDA_WINDOW_HEIGHT:
			g_value_set_int (value, self->height);
			break;

		case TILDA_WINDOW_WIDTH:
			g_value_set_int (value, self->width);
			break;

		case TILDA_WINDOW_X_POSITION:
			g_value_set_int (value, self->x_position);
			break;

		case TILDA_WINDOW_Y_POSITION:
			g_value_set_int (value, self->y_position);
			break;

		case TILDA_WINDOW_TAB_POSITION:
			g_value_set_int (value, self->tab_position);
			break;

		case TILDA_WINDOW_ANIMATION_ORIENTATION:
			g_value_set_int (value, self->animation_orientation);
			break;

		case TILDA_WINDOW_ANIMATION_DELAY:
			g_value_set_int (value, self->animation_delay);
			break;

		case TILDA_WINDOW_KEEP_ABOVE:
			g_value_set_boolean (value, self->keep_above);
			break;

		case TILDA_WINDOW_SKIP_TASKBAR_HINT:
			g_value_set_boolean (value, self->skip_taskbar_hint);
			break;

		case TILDA_WINDOW_STICK:
			g_value_set_boolean (value, self->stick);
			break;

		case TILDA_WINDOW_HIDDEN_AT_START:
			g_value_set_boolean (value, self->hidden_at_start);
			break;

		case TILDA_WINDOW_CENTERED_HORIZONTALLY:
			g_value_set_boolean (value, self->centered_horizontally);
			break;

		case TILDA_WINDOW_CENTERED_VERTICALLY:
			g_value_set_boolean (value, self->centered_vertically);
			break;

		case TILDA_WINDOW_HAVE_REAL_TRANSPARENCY:
			g_value_set_boolean (value, self->have_real_transparency);
			break;

		default:
			/* We don't have this property */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static GObject *
tilda_window_constructor (GType                  type,
						  guint                  n_construct_properties,
						  GObjectConstructParam *construct_properties)
{
	debug_enter ();

	GObject *obj;
	TildaWindow *self;

	/* Invoke parent constructor */
	TildaWindowClass *klass;
	klass = TILDA_WINDOW_CLASS (g_type_class_peek (TILDA_TYPE_WINDOW));
	obj = parent_class->constructor (type,
									 n_construct_properties,
									 construct_properties);

	/* Do other stuff here. The object is ready to go now, and all
	 * ctor properties have been set.
	 */
	self = TILDA_WINDOW(obj);

	/* Register this object with DBus */
	tilda_window_dbus_register_object (self);

	/* Try to set up real transparency */
	tilda_window_setup_real_transparency (self);

	gtk_container_add (GTK_CONTAINER(self->window), self->notebook);
	g_object_set (G_OBJECT(self->notebook), "can-focus", FALSE, NULL);
	gtk_widget_show (self->notebook);

	// FIXME: Remove these, and replace with reads from the config system
	gchar *mykey = g_strdup_printf ("F%d", self->number+3); // TERRIBLE HACK
	g_object_set (G_OBJECT(self), "key", mykey, NULL);
	g_free (mykey);
	g_object_set (G_OBJECT(self), "x-position", 0, "y-position", 0, NULL);
	g_object_set (G_OBJECT(self), "height", 400, "width", 1680, NULL);
	g_object_set (G_OBJECT(self), "keep-above", TRUE, "stick", TRUE, NULL);
	g_object_set (G_OBJECT(self), "hidden-at-start", FALSE, NULL);

	gtk_window_set_decorated (GTK_WINDOW(self->window), FALSE);

	// FIXME: It should be configurable how many terms we add at startup
	tilda_window_add_terminal (self);
	tilda_window_add_terminal (self);

	/* Show us if we're ready. If not, just remain hidden. All sub-widgets must
	 * be gtk_widget_show()n by this point. */
	if (!self->hidden_at_start)
	{
		gtk_widget_show (self->window);
		self->state = WINDOW_DOWN;
	}
	else
		self->state = WINDOW_UP;

	return obj;
}

static void
tilda_window_dispose (GObject *obj)
{
	debug_enter ();

	TildaWindow *self = (TildaWindow *) obj;

	/* We don't want to run dispose twice, so just return immediately */
	if (self->dispose_has_run)
		return;

	/*
	 * In dispose, you are supposed to free all types referenced from this
	 * object which might themselves hold a reference to self. Generally,
	 * the most simple solution is to unref all members on which you own a
	 * reference.
	 *
	 * NOTE: See the following for how to deal with GtkObject-derived things:
	 * http://library.gnome.org/devel/gtk/unstable/GtkObject.html
	 */
	g_ptr_array_foreach (self->terms, g_object_unref, NULL);
	gtk_widget_destroy (self->window);

	/* Unbind if we were set */
	if (self->key)
		tomboy_keybinder_unbind (self->key, tilda_window_keybinding_cb);

	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->dispose (obj);
}

static void
tilda_window_finalize (GObject *obj)
{
	debug_enter ();

	TildaWindow *self = (TildaWindow *) obj;

	/*
	 * Here, complete the object's destruction.
	 * You might not need to do much...
	 */
	// TODO: g_free() any primitives here
	g_ptr_array_free (self->terms, TRUE);


	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
tilda_window_class_init (gpointer g_class,
						 gpointer g_class_data)
{
	debug_enter ();

	GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);
	TildaWindowClass *klass = TILDA_WINDOW_CLASS (g_class);
	GParamSpec *pspec;

	/* Hook our functions to this type */
	gobject_class->set_property = tilda_window_set_property;
	gobject_class->get_property = tilda_window_get_property;
	gobject_class->dispose = tilda_window_dispose;
	gobject_class->finalize = tilda_window_finalize;
	gobject_class->constructor = tilda_window_constructor;

	parent_class = g_type_class_peek_parent (klass);

	/* Install all of the properties */
	pspec = g_param_spec_int ("number",
							  _("Window number"),
							  NULL,
							  0,		// min value
							  INT_MAX,	// max value
							  0,		// def value
							  G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_NUMBER,
									 pspec);

	pspec = g_param_spec_pointer ("controller",
								  _("Pointer to window's controlling TildaController"),
								  NULL,
								  G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_CONTROLLER,
									 pspec);

	pspec = g_param_spec_string ("key",
								 _("Window's drop-down keybinding"),
								 NULL,
								 NULL,
								 G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_KEY,
									 pspec);

	pspec = g_param_spec_int ("height",
							  _("Window's height"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_HEIGHT,
									 pspec);

	pspec = g_param_spec_int ("width",
							  _("Window's width"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_WIDTH,
									 pspec);

	pspec = g_param_spec_int ("x-position",
							  _("Window's x position"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_X_POSITION,
									 pspec);

	pspec = g_param_spec_int ("y-position",
							  _("Window's y position"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_Y_POSITION,
									 pspec);

	pspec = g_param_spec_int ("tab-position",
							  _("Position of window's tab bar"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_TAB_POSITION,
									 pspec);

	pspec = g_param_spec_int ("animation-orientation",
							  _("Window's animation orientation"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_ANIMATION_ORIENTATION,
									 pspec);

	pspec = g_param_spec_int ("animation-delay",
							  _("Amount of time in milliseconds between animation intervals"),
							  NULL,
							  0,
							  INT_MAX,
							  0,
							  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_ANIMATION_DELAY,
									 pspec);

	pspec = g_param_spec_boolean ("keep-above",
								  _("Keep this window above all others"),
								  NULL,
								  FALSE,
								  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_KEEP_ABOVE,
									 pspec);

	pspec = g_param_spec_boolean ("skip-taskbar-hint",
								  _("Hide this window in the taskbar if TRUE"),
								  NULL,
								  FALSE,
								  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_SKIP_TASKBAR_HINT,
									 pspec);

	pspec = g_param_spec_boolean ("stick",
								  _("Display this window on all workspaces"),
								  NULL,
								  FALSE,
								  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_STICK,
									 pspec);

	pspec = g_param_spec_boolean ("hidden-at-start",
								  _("Hide the window when it is first created"),
								  NULL,
								  FALSE,
								  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_HIDDEN_AT_START,
									 pspec);

	pspec = g_param_spec_boolean ("centered-horizontally",
								  _("Center the window horizontally"),
								  NULL,
								  FALSE,
								  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_CENTERED_HORIZONTALLY,
									 pspec);

	pspec = g_param_spec_boolean ("centered-vertically",
								  _("Center the window vertically"),
								  NULL,
								  FALSE,
								  G_PARAM_READWRITE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_CENTERED_VERTICALLY,
									 pspec);

	pspec = g_param_spec_boolean ("have-real-transparency",
								  NULL, NULL, FALSE, G_PARAM_READABLE);

	g_object_class_install_property (gobject_class,
									 TILDA_WINDOW_HAVE_REAL_TRANSPARENCY,
									 pspec);

	/* Hook the TildaWindow type into DBus */
	dbus_g_object_type_install_info (tilda_window_get_type(), &dbus_glib_tilda_window_object_info);
}

GType
tilda_window_get_type (void)
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo info = {
			sizeof (TildaWindowClass),
			NULL,	/* base_init */
			NULL,	/* base_finalize */
			tilda_window_class_init,	/* class_init */
			NULL,	/* class_finalize */
			NULL,	/* class_data */
			sizeof (TildaWindow),
			0,		/* n_preallocs */
			tilda_window_instance_init,	/* instance_init */
		};

		type = g_type_register_static (G_TYPE_OBJECT,
									   "TildaWindowType",
									   &info,
									   0);
	}

	return type;
}

/* vim: set ts=4 sts=4 sw=4 noet tw=112: */
